/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package chess;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author petero
 */
public class TextIO {
    static public final String startPosFEN = new String("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");

    /** Parse a FEN string and return a chess Position object. */
    public static final Position readFEN(String fen) throws ChessParseError {
        Position pos = new Position();
        String[] words = fen.split(" ");
        if (words.length < 2) {
            throw new ChessParseError("Too few spaces");
        }
        
        // Piece placement
        int row = 7;
        int col = 0;
        for (int i = 0; i < words[0].length(); i++) {
            char c = words[0].charAt(i);
            switch (c) {
                case '1': col += 1; break;
                case '2': col += 2; break;
                case '3': col += 3; break;
                case '4': col += 4; break;
                case '5': col += 5; break;
                case '6': col += 6; break;
                case '7': col += 7; break;
                case '8': col += 8; break;
                case '/': row--; col = 0; break;
                case 'P': safeSetPiece(pos, col, row, Piece.WPAWN);   col++; break;
                case 'N': safeSetPiece(pos, col, row, Piece.WKNIGHT); col++; break;
		case 'B': safeSetPiece(pos, col, row, Piece.WBISHOP); col++; break;
		case 'R': safeSetPiece(pos, col, row, Piece.WROOK);   col++; break;
		case 'Q': safeSetPiece(pos, col, row, Piece.WQUEEN);  col++; break;
		case 'K': safeSetPiece(pos, col, row, Piece.WKING);   col++; break;
		case 'p': safeSetPiece(pos, col, row, Piece.BPAWN);   col++; break;
		case 'n': safeSetPiece(pos, col, row, Piece.BKNIGHT); col++; break;
		case 'b': safeSetPiece(pos, col, row, Piece.BBISHOP); col++; break;
		case 'r': safeSetPiece(pos, col, row, Piece.BROOK);   col++; break;
		case 'q': safeSetPiece(pos, col, row, Piece.BQUEEN);  col++; break;
		case 'k': safeSetPiece(pos, col, row, Piece.BKING);   col++; break;
                default: throw new ChessParseError("Invalid piece");
            }
        }
        if (words[1].length() == 0) {
            throw new ChessParseError("Invalid side");
        }
        pos.setWhiteMove(words[1].charAt(0) == 'w');

        // Castling rights
        int castleMask = 0;
        if (words.length > 2) {
            for (int i = 0; i < words[2].length(); i++) {
                char c = words[2].charAt(i);
                switch (c) {
                    case 'K':
                        castleMask |= (1 << Position.CastleBits.H1_CASTLE.ordinal());
                        break;
                    case 'Q':
                        castleMask |= (1 << Position.CastleBits.A1_CASTLE.ordinal());
                        break;
                    case 'k':
                        castleMask |= (1 << Position.CastleBits.H8_CASTLE.ordinal());
                        break;
                    case 'q':
                        castleMask |= (1 << Position.CastleBits.A8_CASTLE.ordinal());
                        break;
                    case '-':
                        break;
                    default:
                        throw new ChessParseError("Invalid castling flags");
                }
            }
        }
        pos.setCastleMask(castleMask);

        if (words.length > 3) {
            // En passant target square
            String epString = words[3];
            if (!epString.equals("-")) {
                if (epString.length() < 2) {
                    throw new ChessParseError("Invalid en passant square");
                }
                pos.setEpSquare(getSquare(epString));
            }
        }
        
        if (words.length > 4) {
            pos.halfMoveClock = Integer.parseInt(words[4]);
        }
        if (words.length > 5) {
            pos.fullMoveCounter = Integer.parseInt(words[5]);
        }
        
        // Each side must have exactly one king
        int wKings = 0;
        int bKings = 0;
        for (int x = 0; x < 8; x++) {
            for (int y = 0; y < 8; y++) {
                Piece p = pos.getPiece(Position.getSquare(x, y));
                if (p == Piece.WKING) {
                    wKings++;
                } else if (p == Piece.BKING) {
                    bKings++;
                }
            }
        }
        if (wKings != 1) {
            throw new ChessParseError("White must have exactly one king");
        }
        if (bKings != 1) {
            throw new ChessParseError("Black must have exactly one king");
        }

        // Make sure king can not be captured
        Position pos2 = new Position(pos);
        pos2.setWhiteMove(!pos.isWhiteMove());
        if (MoveGen.inCheck(pos2)) {
            throw new ChessParseError("King capture possible");
        }
        
        // Remove bogus epSquare
        int epSquare = pos.getEpSquare();
        if (epSquare >= 0) {
            List<Move> moves = new MoveGen().pseudoLegalMoves(pos);
            moves = MoveGen.removeIllegal(pos, moves);
            boolean epValid = false;
            for (Move m : moves) {
                if (m.to == epSquare) {
                    if (pos.getPiece(m.from) == (pos.isWhiteMove() ? Piece.WPAWN : Piece.BPAWN)) {
                        epValid = true;
                        break;
                    }
                }
            }
            if (!epValid) {
                pos.setEpSquare(-1);
            }
        }
        
        return pos;
    }
        
    private static void safeSetPiece(Position pos, int col, int row, Piece p) throws ChessParseError {
        if (row < 0) throw new ChessParseError("Too many rows");
        if (col > 7) throw new ChessParseError("Too many columns");
        if ((p == Piece.WPAWN) || (p == Piece.BPAWN)) {
            if ((row == 0) || (row == 7))
                throw new ChessParseError("Pawn on first/last rank");
        }
        pos.setPiece(Position.getSquare(col, row), p);
    }
    
    /** Return a FEN string corresponding to a chess Position object. */
    public static final String toFEN(Position pos) {
        StringBuilder ret = new StringBuilder();
        // Piece placement
        for (int r = 7; r >=0; r--) {
            int numEmpty = 0;
            for (int c = 0; c < 8; c++) {
                Piece p = pos.getPiece(Position.getSquare(c, r));
                if (p == Piece.EMPTY) {
                    numEmpty++;
                } else {
                    if (numEmpty > 0) {
                        ret.append(numEmpty);
                        numEmpty = 0;
                    }
                    switch (p) {
                        case WKING:   ret.append('K'); break;
                        case WQUEEN:  ret.append('Q'); break;
                        case WROOK:   ret.append('R'); break;
                        case WBISHOP: ret.append('B'); break;
                        case WKNIGHT: ret.append('N'); break;
                        case WPAWN:   ret.append('P'); break;
                        case BKING:   ret.append('k'); break;
                        case BQUEEN:  ret.append('q'); break;
                        case BROOK:   ret.append('r'); break;
                        case BBISHOP: ret.append('b'); break;
                        case BKNIGHT: ret.append('n'); break;
                        case BPAWN:   ret.append('p'); break;
                        default: throw new RuntimeException();
                    }
                }
            }
            if (numEmpty > 0) {
                ret.append(numEmpty);
            }
            if (r > 0) {
                ret.append('/');
            }
        }
        ret.append(pos.isWhiteMove() ? " w " : " b ");

        // Castling rights
        boolean anyCastle = false;
        if (pos.h1Castle()) {
            ret.append('K');
            anyCastle = true;
        }
        if (pos.a1Castle()) {
            ret.append('Q');
            anyCastle = true;
        }
        if (pos.h8Castle()) {
            ret.append('k');
            anyCastle = true;
        }
        if (pos.a8Castle()) {
            ret.append('q');
            anyCastle = true;
        }
        if (!anyCastle) {
            ret.append('-');
        }
        
        // En passant target square
        {
            ret.append(' ');
            if (pos.getEpSquare() >= 0) {
                int x = Position.getX(pos.getEpSquare());
                int y = Position.getY(pos.getEpSquare());
                ret.append((char)(x + 'a'));
                ret.append((char)(y + '1'));
            } else {
                ret.append('-');
            }
        }

        // Move counters
        ret.append(' ');
        ret.append(pos.halfMoveClock);
        ret.append(' ');
        ret.append(pos.fullMoveCounter);

        return ret.toString();
    }
    
    /**
     * Convert a chess move to human readable form.
     * @param pos      The chess position.
     * @param move     The executed move.
     * @param longForm If true, use long notation, eg Ng1-f3.
     *                 Otherwise, use short notation, eg Nf3
     */
    public static final String moveToString(Position pos, Move move, boolean longForm) {
        StringBuilder ret = new StringBuilder();
        int wKingOrigPos = Position.getSquare(4, 0);
        int bKingOrigPos = Position.getSquare(4, 7);
        if (move.from == wKingOrigPos && pos.getPiece(wKingOrigPos) == Piece.WKING) {
            // Check white castle
            if (move.to == Position.getSquare(6, 0)) {
                    ret.append("O-O");
            } else if (move.to == Position.getSquare(2, 0)) {
                ret.append("O-O-O");
            }
        } else if (move.from == bKingOrigPos && pos.getPiece(bKingOrigPos) == Piece.BKING) {
            // Check white castle
            if (move.to == Position.getSquare(6, 7)) {
                ret.append("O-O");
            } else if (move.to == Position.getSquare(2, 7)) {
                ret.append("O-O-O");
            }
        }
        if (ret.length() == 0) {
            Piece p = pos.getPiece(move.from);
            ret.append(pieceToChar(p));
            int x1 = Position.getX(move.from);
            int y1 = Position.getY(move.from);
            int x2 = Position.getX(move.to);
            int y2 = Position.getY(move.to);
            if (longForm) {
                ret.append((char)(x1 + 'a'));
                ret.append((char) (y1 + '1'));
                ret.append(isCapture(pos, move) ? 'x' : '-');
            } else {
                if (p == (pos.isWhiteMove() ? Piece.WPAWN : Piece.BPAWN)) {
                    if (isCapture(pos, move)) {
                        ret.append((char) (x1 + 'a'));
                    }
                } else {
                    int numSameTarget = 0;
                    int numSameFile = 0;
                    int numSameRow = 0;
                    List<Move> moves = new MoveGen().pseudoLegalMoves(pos);
                    moves = MoveGen.removeIllegal(pos, moves);
                    for (Move m : moves) {
                        if ((pos.getPiece(m.from) == p) && (m.to == move.to)) {
                            numSameTarget++;
                            if (Position.getX(m.from) == x1)
                                numSameFile++;
                            if (Position.getY(m.from) == y1)
                                numSameRow++;
                        }
                    }
                    if (numSameTarget < 2) {
                        // No file/row info needed
                    } else if (numSameFile < 2) {
                        ret.append((char) (x1 + 'a'));   // Only file info needed
                    } else if (numSameRow < 2) {
                        ret.append((char) (y1 + '1'));   // Only row info needed
                    } else {
                        ret.append((char) (x1 + 'a'));   // File and row info needed
                        ret.append((char) (y1 + '1'));
                    }
                }
                if (isCapture(pos, move)) {
                    ret.append('x');
                }
            }
            ret.append((char) (x2 + 'a'));
            ret.append((char) (y2 + '1'));
            if (move.promoteTo != Piece.EMPTY) {
                ret.append(pieceToChar(move.promoteTo));
            }
        }
        UndoInfo ui = new UndoInfo();
        pos.makeMove(move, ui);
        if (MoveGen.inCheck(pos)) {
            List<Move> moves = new MoveGen().pseudoLegalMoves(pos);
            moves = MoveGen.removeIllegal(pos, moves);
            if (moves.size() == 0) {
                ret.append('#');
            } else {
                ret.append('+');
            }
        }
        pos.unMakeMove(move, ui);

        return ret.toString();
    }

    private static boolean isCapture(Position pos, Move move) {
        if (pos.getPiece(move.to) == Piece.EMPTY) {
            Piece p = pos.getPiece(move.from);
            if ((p == (pos.isWhiteMove() ? Piece.WPAWN : Piece.BPAWN)) && (move.to == pos.getEpSquare())) {
                return true;
            } else {
                return false;
            }
        } else {
            return true;
        }
    }

    /**
     * Convert a chess move string to a Move object.
     * Any prefix of the string representation of a valid move counts as a legal move string,
     * as long as the string only matches one valid move.
     */
    public static final Move stringToMove(Position pos, String strMove) {
        strMove = strMove.replaceAll("=", "");
        Move move = null;
        if (strMove.length() == 0)
            return move;
        List<Move> moves = new MoveGen().pseudoLegalMoves(pos);
        moves = MoveGen.removeIllegal(pos, moves);
        {
            char lastChar = strMove.charAt(strMove.length() - 1);
            if ((lastChar == '#') || (lastChar == '+')) {
                List<Move> subMoves = new ArrayList<Move>();
                for (Move m : moves) {
                    String str1 = TextIO.moveToString(pos, m, true);
                    if (str1.charAt(str1.length() - 1) == lastChar) {
                        subMoves.add(m);
                    }
                }
                moves = subMoves;
                strMove = normalizeMoveString(strMove);
            }
        }

        for (int i = 0; i < 2; i++) {
            // Search for full match
            for (Move m : moves) {
                String str1 = normalizeMoveString(TextIO.moveToString(pos, m, true));
                String str2 = normalizeMoveString(TextIO.moveToString(pos, m, false));
                if (i == 0) {
                    if (strMove.equals(str1) || strMove.equals(str2)) {
                        return m;
                    }
                } else {
                    if (strMove.toLowerCase().equals(str1.toLowerCase()) ||
                            strMove.toLowerCase().equals(str2.toLowerCase())) {
                        return m;
                    }
                }
            }
        }
        
        for (int i = 0; i < 2; i++) {
            // Search for unique substring match
            for (Move m : moves) {
                String str1 = normalizeMoveString(TextIO.moveToString(pos, m, true));
                String str2 = normalizeMoveString(TextIO.moveToString(pos, m, false));
                boolean match;
                if (i == 0) {
                    match = (str1.startsWith(strMove) || str2.startsWith(strMove));
                } else {
                    match = (str1.toLowerCase().startsWith(strMove.toLowerCase()) ||
                            str2.toLowerCase().startsWith(strMove.toLowerCase()));
                }
                if (match) {
                    if (move != null) {
                        return null; // More than one match, not ok
                    } else {
                        move = m;
                    }
                }
            }
            if (move != null)
                return move;
        }
        return move;
    }

    /**
     * Convert a string, such as "e4" to a square number.
     * @return The square number, or -1 if not a legal square.
     */
    public static int getSquare(String s) {
        int x = s.charAt(0) - 'a';
        int y = s.charAt(1) - '1';
        if ((x < 0) || (x > 7) || (y < 0) || (y > 7))
            return -1;
        return Position.getSquare(x, y);
    }

    /**
     * Convert a square number to a string, such as "e4".
     */
    public static String squareToString(int square) {
        StringBuilder ret = new StringBuilder();
        int x = Position.getX(square);
        int y = Position.getY(square);
        ret.append((char) (x + 'a'));
        ret.append((char) (y + '1'));
        return ret.toString();
    }

    /**
     * Create an ascii representation of a position.
     */
    public static final String asciiBoard(Position pos) {
        StringBuilder ret = new StringBuilder(400);
        String nl = String.format("%n");
        ret.append("    +----+----+----+----+----+----+----+----+"); ret.append(nl);
        for (int y = 7; y >= 0; y--) {
            ret.append("    |");
            for (int x = 0; x < 8; x++) {
                ret.append(' ');
                Piece p = pos.getPiece(Position.getSquare(x, y));
                if (p == Piece.EMPTY) {
                    boolean dark = Position.darkSquare(x, y);
                    ret.append(dark ? ".. |" : "   |");
                } else {
                    ret.append(p.isWhite() ? ' ' : '*');
                    String pieceName = pieceToChar(p);
                    if (pieceName.length() == 0)
                        pieceName = "P";
                    ret.append(pieceName);
                    ret.append(" |");
                }
            }
            ret.append(nl);
            ret.append("    +----+----+----+----+----+----+----+----+");
            ret.append(nl);
        }
        return ret.toString();
    }

    /**
     * Convert move string to lower case and remove special check/mate symbols.
     */
    private static String normalizeMoveString(String str) {
        if (str.length() > 0) {
            char lastChar = str.charAt(str.length() - 1);
            if ((lastChar == '#') || (lastChar == '+')) {
                str = str.substring(0, str.length() - 1);
            }
        }
        return str;
    }
    
    private final static String pieceToChar(Piece p) {
        switch (p) {
            case WQUEEN:  case BQUEEN:  return "Q";
            case WROOK:   case BROOK:   return "R";
            case WBISHOP: case BBISHOP: return "B";
            case WKNIGHT: case BKNIGHT: return "N";
            case WKING:   case BKING:   return "K";
        }
        return "";
    }
}
