/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package uci;

import chess.Book;
import chess.Move;
import chess.MoveGen;
import chess.Position;
import chess.Search;
import chess.TextIO;
import chess.TranspositionTable;
import chess.TranspositionTable.TTEntry;
import chess.UndoInfo;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

/**
 * Control the search thread.
 * @author petero
 */
public class EngineControl {
    PrintStream os;

    Thread engineThread;
    private final Object threadMutex;
    Search sc;
    TranspositionTable tt;
    MoveGen moveGen;

    Position pos;
    List<Long> posHashList;
    boolean ponder;
    boolean infinite;

    int timeLimit;
    int maxDepth;
    List<Move> searchMoves;

    // Options
    int hashSizeMB = 16;
    boolean ownBook = false;
    boolean analyseMode = false;

    /**
     * This class is responsible for sending "info" strings during search.
     */
    static class SearchListener implements Search.Listener {
        PrintStream os;
        
        SearchListener(PrintStream os) {
            this.os = os;
        }

        public void notifyDepth(int depth) {
            os.printf("info depth %d%n", depth);
        }

        public void notifyCurrMove(Move m, int moveNr) {
            os.printf("info currmove %s currmovenumber %d%n", moveToString(m), moveNr);
        }

        public void notifyPV(int depth, int score, int time, int nodes, int nps, boolean isMate,
                boolean upperBound, boolean lowerBound, List<Move> pv) {
            StringBuilder pvBuf = new StringBuilder();
            for (Move m : pv) {
                pvBuf.append(" ");
                pvBuf.append(moveToString(m));
            }
            String bound = "";
            if (upperBound) {
                bound = " upperbound";
            } else if (lowerBound) {
                bound = " lowerbound";
            }
            os.printf("info depth %d score %s %d%s time %d nodes %d nps %d pv%s%n",
                    depth, isMate ? "mate" : "cp", score, bound, time, nodes, nps, pvBuf.toString());
        }

        public void notifyStats(int nodes, int nps, int time) {
            os.printf("info nodes %d nps %d time %d%n", nodes, nps, time);
        }
    }

    public EngineControl(PrintStream os) {
        this.os = os;
        threadMutex = new Object();
        setupTT();
        moveGen = new MoveGen();
    }

    final public void startSearch(Position pos, List<Move> moves, SearchParams sPar) {
        setupPosition(new Position(pos), moves);
        computeTimeLimit(sPar);
        ponder = false;
        infinite = (timeLimit < 0) && (maxDepth < 0);
        startThread(timeLimit, maxDepth);
        searchMoves = sPar.searchMoves;
    }

    final public void startPonder(Position pos, List<Move> moves, SearchParams sPar) {
        setupPosition(new Position(pos), moves);
        computeTimeLimit(sPar);
        ponder = true;
        infinite = false;
        startThread(-1, -1);
    }

    final public void ponderHit() {
        Search mySearch;
        synchronized (threadMutex) {
            mySearch = sc;
        }
        if (mySearch != null) {
            mySearch.timeLimit(timeLimit);
        }
        infinite = (timeLimit < 0) && (maxDepth < 0);
        ponder = false;
    }

    final public void stopSearch() {
        stopThread();
    }

    /**
     * Compute thinking time for current search.
     */
    final public void computeTimeLimit(SearchParams sPar) {
        timeLimit = -1;
        maxDepth = -1;
        if (sPar.infinite) {
            timeLimit = -1;
            maxDepth = -1;
        } else if (sPar.depth > 0) {
            maxDepth = sPar.depth;
        } else if (sPar.mate > 0) {
            maxDepth = sPar.mate * 2 - 1;
        } else if (sPar.moveTime > 0) {
            timeLimit = sPar.moveTime;
        } else if (sPar.nodes > 0) {
            timeLimit = sPar.nodes / 75; // FIXME!!! Implement for real
        } else {
            int moves = sPar.movesToGo;
            if (moves == 0) {
                moves = 40; // Assume 40 more moves until end of game
            }
            boolean white = pos.isWhiteMove();
            int time = white ? sPar.wTime : sPar.bTime;
            int inc  = white ? sPar.wInc : sPar.bInc;
            timeLimit = (time + inc * (moves - 1) - 1000) / moves;
            timeLimit = Math.min(timeLimit, time - 1000); // Leave at least 1s on the clock
            timeLimit = Math.max(0, timeLimit);           // But can't use negative time
        }
    }

    final public void startThread(int timeLimit, final int maxDepth) {
        sc = new Search(pos, posHashList, tt);
        sc.setListener(new SearchListener(os));
        List<Move> moves = moveGen.pseudoLegalMoves(pos);
        moves = MoveGen.removeIllegal(pos, moves);
        if ((searchMoves != null) && (searchMoves.size() > 0)) {
            moves.retainAll(searchMoves);
        }
        final List<Move> srchMoves = moves;
        if ((srchMoves.size() <= 1) && !infinite) {
            timeLimit = 0;
        }
        final int srchTimeLimit = timeLimit;
        engineThread = new Thread(new Runnable() {
            public void run() {
                Move m = null;
                if (ownBook && !analyseMode) {
                    Book book = new Book(false);
                    m = book.getBookMove(pos);
                }
                if (m == null) {
                    m = sc.iterativeDeepening(srchMoves, srchTimeLimit, maxDepth, false);
                }
                while (ponder || infinite) {
                    // We should not respond until told to do so. Just wait until
                    // we are allowed to respond.
                    try {
                        Thread.sleep(10);
                    } catch (InterruptedException ex) {
                        break;
                    }
                }
                Move ponderMove = getPonderMove(pos, m);
                if (ponderMove != null) {
                    os.printf("bestmove %s ponder %s%n", moveToString(m), moveToString(ponderMove));
                } else {
                    os.printf("bestmove %s%n", moveToString(m));
                }
                synchronized (threadMutex) {
                    engineThread = null;
                    sc = null;
                }
            }
        });
        engineThread.start();
    }

    final public void stopThread() {
        Thread myThread;
        Search mySearch;
        synchronized (threadMutex) {
            myThread = engineThread;
            mySearch = sc;
        }
        if (myThread != null) {
            mySearch.timeLimit(0);
            infinite = false;
            ponder = false;
            try {
                myThread.join();
            } catch (InterruptedException ex) {
                throw new RuntimeException();
            }
        }
    }


    private final void setupTT() {
        int nEntries = hashSizeMB > 0 ? hashSizeMB * (1 << 20) / 24 : 1024;
        int logSize = (int) Math.floor(Math.log(nEntries) / Math.log(2));
        tt = new TranspositionTable(logSize);
    }

    final void setupPosition(Position pos, List<Move> moves) {
        UndoInfo ui = new UndoInfo();
        posHashList = new ArrayList<Long>();
        for (Move m : moves) {
            posHashList.add(pos.zobristHash());
            pos.makeMove(m, ui);
        }
        this.pos = pos;
    }

    /**
     * Try to find a move to ponder from the transposition table.
     */
    final Move getPonderMove(Position pos, Move m) {
        Move ret = null;
        UndoInfo ui = new UndoInfo();
        pos.makeMove(m, ui);
        TTEntry ent = tt.probe(pos.historyHash());
        if (ent.type != TTEntry.T_EMPTY) {
            ret = ent.getMove();
            List<Move> moves = moveGen.pseudoLegalMoves(pos);
            moves = MoveGen.removeIllegal(pos, moves);
            if (!moves.contains(ret)) {
                ret = null;
            }
        }
        pos.unMakeMove(m, ui);
        return ret;
    }

    static final String moveToString(Move m) {
        String ret = TextIO.squareToString(m.from);
        ret += TextIO.squareToString(m.to);
        switch (m.promoteTo) {
            case WQUEEN:
            case BQUEEN:
                ret += "q";
                break;
            case WROOK:
            case BROOK:
                ret += "r";
                break;
            case WBISHOP:
            case BBISHOP:
                ret += "b";
                break;
            case WKNIGHT:
            case BKNIGHT:
                ret += "n";
                break;
            default:
                break;
        }
        return ret;
    }

    static void printOptions(PrintStream os) {
        os.printf("option name Hash type spin default 16 min 1 max 2048%n");
        os.printf("option name OwnBook type check default false%n");
        os.printf("option name Ponder type check default true%n");
        os.printf("option name UCI_AnalyseMode type check default false%n");
        os.printf("option name UCI_EngineAbout type string default CuckooChess by Peter Osterlund, see http://web.telia.com/~u89404340/javachess/index.html%n");
    }

    final void setOption(String optionName, String optionValue) {
        try {
            if (optionName.equals("hash")) {
                hashSizeMB = Integer.parseInt(optionValue);
                setupTT();
            } else if (optionName.equals("ownbook")) {
                ownBook = Boolean.parseBoolean(optionValue);
            } else if (optionName.equals("ponder")) {
                // Do nothing for now. This could be used for changing time management.
            } else if (optionName.equals("uci_analysemode")) {
                analyseMode = Boolean.parseBoolean(optionValue);
            }
        } catch (NumberFormatException nfe) {
        }
    }
}
